/****************************************************************************
 Creation 17/04/2001                                               by Sylvain


                           METAFILE.C
                           ----------


*****************************************************************************

         Ce programme assure :

  minimum API for :
  -reading wmf files
  -saving emf to bmp's file

  (Original sources come from http://www.codeguru.com/ --> metafile_example.zip )
****************************************************************************/
#include <windows.h>
#include <stdio.h>
#include <io.h> 
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h> 
#include "metafile.h"





// Placeable metafile data definitions 
typedef struct tagOLDRECT
{
    short left;
    short top;
    short right;
    short bottom;
} OLDRECT;

// Placeable metafile header 
typedef struct {
        DWORD   key;
        WORD	hmf;
        OLDRECT	bbox;
        WORD    inch;
        DWORD   reserved;
        WORD    checksum;
} ALDUSMFHEADER;


#define	METAFILE_VERSION	1
#define	ALDUSKEY		0x9AC6CDD7
#define	ALDUSMFHEADERSIZE	22	// Avoid sizeof is struct alignment >1



/*---------------------------------------------------------------------------
**
**        WMF2BMP
**---------------------------------------------------------------------------
**
**  DESCRIPTION
**        --Wmf2Bmp--
**  ARGUMENTS
**        Argument1:  char * pszWMFFile
**        Argument2: char * pszBmpFile
**  RETOUR/RESULTAT
**        BOOL 
**  DIAGNOSTIC
**  IMPLEMENTATION
**  REMARQUE
**  Rev 17/04/2001 
**--------------------------------------------------------------------------*/
BOOL Wmf2Bmp( char * pszWMFFile, char * pszBmpFile )
{
    HDC hDC;
    HENHMETAFILE hemf; RECT rect;
    BOOL bRet;
	
    hDC = GetDC( NULL);    
    hemf = OpenMetafile( pszWMFFile, &rect );
    if( hemf != NULL )
    {
	bRet = SaveMetafileToBitmap( hDC, hemf, &rect, pszBmpFile );
    }
    else
    {
	bRet = FALSE;
    }
    ReleaseDC( NULL, hDC );
    if( hemf )
	DeleteEnhMetaFile(hemf); 
    return bRet;
}



/*---------------------------------------------------------------------------
**
**        OPENMETAFILE
**---------------------------------------------------------------------------
**
**  DESCRIPTION
**        --OpenMetafile--
**  ARGUMENTS
**        Argument1:  char * pszFile	[IN]	WMF or EMF file
**        Argument2: RECT * pRect	[OUT]	size
**  RETOUR/RESULTAT
**        HENHMETAFILE	EMF handle
**  DIAGNOSTIC
**  IMPLEMENTATION
**  REMARQUE
**  Rev 17/04/2001 
**--------------------------------------------------------------------------*/
HENHMETAFILE OpenMetafile( char * pszFile, RECT * pRect )
{
    DWORD   		dwIsAldus;
    METAHEADER		mfHeader;
    DWORD    		dwSize;
    DWORD		seekpos;
    int hFile; int cbRead;
    ALDUSMFHEADER m_aldusMFHeader;
    HENHMETAFILE hemf = NULL;
    char * lpMFBits;

    hFile = open( pszFile, O_RDONLY | O_BINARY );
    if( hFile == -1 )
	return NULL;
    
    if (read( hFile, &dwIsAldus, sizeof(dwIsAldus)) != sizeof(dwIsAldus))
    {
	close( hFile );
	return NULL;
    }
    
    if (dwIsAldus != ALDUSKEY) {
	// A windows metafile, not a placeable wmf 
	seekpos = 0;
	ZeroMemory(&m_aldusMFHeader, sizeof(m_aldusMFHeader));
    }
    else {  
	// This is a placeable metafile 
	// Convert the placeable format into something that can
	// be used with GDI metafile functions 
	lseek( hFile, 0, SEEK_SET);
	if (read( hFile, &m_aldusMFHeader, ALDUSMFHEADERSIZE) != ALDUSMFHEADERSIZE)
	{
	    close( hFile );
	    return NULL;
	}
	seekpos = ALDUSMFHEADERSIZE;
    }
    lseek( hFile, seekpos, SEEK_SET);
    // Read the metafile header 
    if (read( hFile, &mfHeader, sizeof(mfHeader)) != sizeof(mfHeader))
    {
	close( hFile );
	return NULL;
    }
    // At this point we have a metafile header regardless of whether
    // the metafile was a windows metafile or a placeable metafile
    // so check to see if it is valid.  There is really no good
    // way to do this so just make sure that the mtType is either
    // 1 or 2 (memory or disk file) 
    if ((mfHeader.mtType != 1) && (mfHeader.mtType != 2)) 
    {
	close( hFile );
	return NULL;
    }
    // Allocate memory for the metafile bits 
    dwSize = mfHeader.mtSize * 2;
    lpMFBits = (char *)malloc( dwSize );
    if (lpMFBits == NULL)
    {
	close( hFile );
	return NULL;
    }
    // Seek to the metafile bits 
    lseek( hFile, seekpos, SEEK_SET);
    // Read metafile bits and create
    // the enhanced metafile
    cbRead = read( hFile, lpMFBits, dwSize);
    if ( cbRead == (int)dwSize )
	hemf = SetWinMetaFileBits(dwSize, lpMFBits, NULL, NULL);
    // Done with the bit data	
    free( lpMFBits );
    close( hFile );
    
    // Set alignment
    // We can only do this if we can
    // get the size of the metafile
    // This should be possible for
    // metafiles with Aldus headers
    pRect->left = 0;
    pRect->top = 0;
    pRect->right = 400;
    pRect->bottom = 0;
    if (m_aldusMFHeader.inch) {
	// Get the logical size of the metafile
	pRect->right = 254 * (m_aldusMFHeader.bbox.right - m_aldusMFHeader.bbox.left) 
			/ m_aldusMFHeader.inch;
	pRect->bottom = 254 * (m_aldusMFHeader.bbox.bottom - m_aldusMFHeader.bbox.top) 
			/ m_aldusMFHeader.inch;
    }
    return hemf;
}




///////////////////////////////////////////
static BOOL SaveBitmapFile(HDC hDC, HBITMAP hBmp, PSTR pszFileName)
{
    int         hFile;
    OFSTRUCT    ofReOpenBuff;
    HBITMAP     hTmpBmp, hBmpOld;
    BOOL        bSuccess;
    BITMAPFILEHEADER    bfh;
    PBITMAPINFO pbmi; BOOL fpbmiAllocate = FALSE;
    PBYTE       pBits;
    BITMAPINFO  bmi;
    PBYTE pjTmp, pjTmpBmi;
    ULONG sizBMI;


    bSuccess = TRUE;
    if (!hBmp) {
        return FALSE;
    }

    //
    // Let the graphics engine to retrieve the dimension of the bitmap for us
    // GetDIBits uses the size to determine if it's BITMAPCOREINFO or BITMAPINFO
    // if BitCount != 0, color table will be retrieved
    //
    bmi.bmiHeader.biSize = 0x28;              // GDI need this to work
    bmi.bmiHeader.biBitCount = 0;             // don't get the color table
    if ((GetDIBits(hDC, hBmp, 0, 0, (LPSTR)NULL, &bmi, DIB_RGB_COLORS)) == 0) {
       return FALSE;
    }

    //
    // Now that we know the size of the image, alloc enough memory to retrieve
    // the actual bits
    //
    if ((pBits = (PBYTE)GlobalAlloc(GMEM_FIXED | GMEM_ZEROINIT,
                bmi.bmiHeader.biSizeImage)) == NULL) {
        return FALSE;
    }

    //
    // Note: 24 bits per pixel has no color table.  So, we don't have to
    // allocate memory for retrieving that.  Otherwise, we do.
    //
    pbmi = &bmi;                                      // assume no color table

    switch (bmi.bmiHeader.biBitCount) {
        case 24:                                      // has color table
            sizBMI = sizeof(BITMAPINFOHEADER);
            break;
        case 16:
        case 32:
            sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(DWORD)*3;
            break;
        default:
            sizBMI = sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*(1<<bmi.bmiHeader.biBitCount);
            break;

    }

    //
    // Allocate memory for color table if it is not 24bpp...
    //
    if (sizBMI != sizeof(BITMAPINFOHEADER)) {
        ULONG       sizTmp;
        //
        // I need more memory for the color table
        //
        if ((pbmi = (PBITMAPINFO)GlobalAlloc(GMEM_FIXED | GMEM_ZEROINIT, sizBMI )) == NULL) {
            bSuccess = FALSE;
            goto ErrExit1;
        }
	fpbmiAllocate = TRUE;
        //
        // Now that we've a bigger chunk of memory, let's copy the Bitmap
        // info header data over
        //
        pjTmp = (PBYTE)pbmi;
        pjTmpBmi = (PBYTE)&bmi;
        sizTmp = sizeof(BITMAPINFOHEADER);

        while(sizTmp--)
        {
            *(((PBYTE)pjTmp)++) = *((pjTmpBmi)++);
        }
    }

    //
    // Let's open the file and get ready for writing
    //
    if ((hFile = OpenFile(pszFileName, (LPOFSTRUCT)&ofReOpenBuff,
                 OF_CREATE | OF_WRITE)) == -1) {
        goto ErrExit2;
    }

    //
    // But first, fill in the info for the BitmapFileHeader
    //
    bfh.bfType = 0x4D42;                            // 'BM'
    bfh.bfSize = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+sizBMI+
        pbmi->bmiHeader.biSizeImage;
    bfh.bfReserved1 =
    bfh.bfReserved2 = 0;
    bfh.bfOffBits = sizeof(BITMAPFILEHEADER)+sizBMI;

    //
    // Write out the file header now
    //
    if (_lwrite(hFile, (LPSTR)&bfh, sizeof(BITMAPFILEHEADER)) == -1) {
        bSuccess = FALSE;
        goto ErrExit3;
    }

    //
    // Bitmap can't be selected into a DC when calling GetDIBits
    // Assume that the hDC is the DC where the bitmap would have been selected
    // if indeed it has been selected
    //
    if (hTmpBmp = CreateCompatibleBitmap(hDC, pbmi->bmiHeader.biWidth, pbmi->bmiHeader.biHeight)) {
        hBmpOld = SelectObject(hDC, hTmpBmp);
        if ((GetDIBits(hDC, hBmp, 0, pbmi->bmiHeader.biHeight, (LPSTR)pBits, pbmi, DIB_RGB_COLORS))==0){
            bSuccess = FALSE;
            goto ErrExit4;
        }
    } else {
        bSuccess = FALSE;
        goto ErrExit3;
    }

    //
    // Now write out the BitmapInfoHeader and color table, if any
    //
    if (_lwrite(hFile, (LPSTR)pbmi, sizBMI) == -1) {
        bSuccess = FALSE;
        goto ErrExit4;
    }

    //
    // write the bits also
    //
    if (_lwrite(hFile, (LPSTR)pBits, pbmi->bmiHeader.biSizeImage) == -1) {
        bSuccess = FALSE;
        goto ErrExit4;
    }


ErrExit4:
    SelectObject(hDC, hBmpOld);
    DeleteObject(hTmpBmp);
ErrExit3:
    _lclose(hFile);
ErrExit2:
    if( fpbmiAllocate )
	GlobalFree(pbmi);
ErrExit1:
    GlobalFree(pBits);
    return bSuccess;
}











/*---------------------------------------------------------------------------
**
**        SAVEMETAFILETOBITMAP
**---------------------------------------------------------------------------
**
**  DESCRIPTION
**        --SaveMetafileToBitmap--
**  ARGUMENTS
**        Argument1:  HDC hDC		    [IN]
**        Argument2: HENHMETAFILE hemf	    [IN]
**        Argument3: RECT * pRect	    [IN]
**        Argument4: char * pszBitmapFile   [IN]
**  RETOUR/RESULTAT
**        BOOL TRUE ok, FALSE, bad
**  DIAGNOSTIC
**  IMPLEMENTATION
**  REMARQUE
**  Rev 17/04/2001 
**--------------------------------------------------------------------------*/
BOOL SaveMetafileToBitmap( HDC hDC, HENHMETAFILE hemf, RECT * pRect, char * pszBitmapFile )
{
    HBITMAP hBmp;
    HDC hDCMem; 
    HGDIOBJ hOld;
    BOOL bRet;
    
    // Create the bitmap
    hBmp = CreateCompatibleBitmap( hDC, pRect->right, pRect->bottom );
    if( hBmp == NULL )
	return FALSE;
    hDCMem = CreateCompatibleDC(hDC);
    if( hDCMem == NULL )
	DeleteObject( hBmp );

    // Selct bitmap in memory
    hOld = SelectObject(hDCMem, hBmp );
    PatBlt(hDCMem, 0, 0, pRect->right, pRect->bottom, PATCOPY);

    // Play the file in memory
    PlayEnhMetaFile(hDCMem, hemf, pRect); 
    
    // Save the bitmap
    bRet = SaveBitmapFile(hDCMem, hBmp, pszBitmapFile );
    
    // Restore and free memory
    SelectObject(hDCMem, hOld);
    DeleteDC( hDCMem );
    DeleteObject( hBmp );
    return bRet;
}




